#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>

static char const prog[]="delayrun";

static void alarm_handler(int sig) {
  (void)sig;
}

int main(int argc, char** argv) {
  struct sigaction action={ 0 };
  sigset_t alarmset;
  sigset_t blockset;
  unsigned long delay=0;
  char const* x;

  /* set up sigset containing SIGALRM */
  if (sigemptyset(&alarmset)!=0)
    strerr_die6sys(100, prog, fatal, err_unable, err_empty, err_sigset, err_);
  if (sigaddset(&alarmset, SIGALRM)!=0)
    strerr_die6sys(100, prog, fatal, err_unable, err_addto, err_sigset, err_);

  /* block SIGALRM ASAP */
  if (sigprocmask(SIG_BLOCK, &alarmset, &blockset)!=0)
    strerr_die5sys(100, prog, fatal, err_unable, err_sigmask, err_);

  /* initialize SIGALRM action, for catching or ignoring later */
  if (sigemptyset(&action.sa_mask)!=0)
    strerr_die6sys(100, prog, fatal, err_unable, err_empty, err_sigset, err_);
  action.sa_flags=0;

  /* scan delay time */
  x=env_get(var_delay+1);
  if (x!=null) {
    unsigned int const len=scan_ulong(x, &delay);
    if (len==0 || x[len]!='\0')
      strerr_die4x(111, prog, fatal, var_delay, err_malformed);
    if (delay>LONG_MAX)
      strerr_die5x(100, prog, fatal, var_delay, err_oflow, t_long);
  }

  if (x==null || delay!=0) { /* do we need to sleep? */
    { /* schedule SIGALRM */
      struct itimerval timer;
      timer.it_interval.tv_sec=0;
      timer.it_interval.tv_usec=0;
      timer.it_value.tv_sec=delay;
      timer.it_value.tv_usec=0;
      if (setitimer(ITIMER_REAL, &timer, null)!=0)
        strerr_die5sys(100, prog, fatal, err_unable, err_timer, err_);
    }

    /* catch SIGALRM */
    action.sa_handler=alarm_handler;
    if (sigaction(SIGALRM, &action, null)!=0)
      strerr_die5sys(100, prog, fatal, err_unable, err_sigact, err_);

    /* sleep */
    if (sigdelset(&blockset, SIGALRM)!=0)
      strerr_die6sys(100, prog, fatal, err_unable, err_delfrom, err_sigset,
                     err_);
    if (sigsuspend(&blockset)!=0 && errno!=EINTR)
      strerr_die5sys(100, prog, fatal, err_unable, err_suspend, err_);
  }

  /* ignore and unblock SIGALRM */
  action.sa_handler=SIG_IGN;
  if (sigaction(SIGALRM, &action, null)!=0)
    strerr_die5sys(100, prog, fatal, err_unable, err_sigact, err_);
  if (sigprocmask(SIG_UNBLOCK, &alarmset, null)!=0)
    strerr_die5sys(100, prog, fatal, err_unable, err_sigmask, err_);

  if (argc<2) _exit(0);
  ++argv;
  pathexec_run(argv[0], (char const**)argv, (char const**)environ);
  strerr_die6sys(errstat(errno), prog, fatal, err_unable, err_exec, argv[0],
                 err_);
}
